import { IterableSerializer, ObjectSerializer, Serializable, Serializer, serializeWith } from 'serde/ser'

class JSONObjectSerializer implements ObjectSerializer<void> {
  private ser: JSONSerializer
  private first: boolean = true

  constructor(serializer: JSONSerializer) {
    this.ser = serializer
    serializer.write('{')
  }

  serializeKey<U extends Serializable>(key: U): void {
    if (!this.first) {
      this.ser.write(',')
    } else {
      this.first = false
    }

    serializeWith(this.ser, key)
    this.ser.write(':')
  }

  serializeValue<U extends Serializable>(value: U): void {
    serializeWith(this.ser, value)
  }

  end(): void {
    this.ser.write('}')
  }
}

class JSONIterableSerializer implements IterableSerializer<void> {
  private ser: JSONSerializer
  private first: boolean = true

  constructor(serializer: JSONSerializer) {
    this.ser = serializer
    serializer.write('[')
  }

  serializeElement<U extends Serializable>(element: U): void {
    if (!this.first) {
      this.ser.write(',')
    } else {
      this.first = false
    }

    serializeWith(this.ser, element)
  }

  end(): void {
    this.ser.write(']')
  }
}

class JSONClassSerializer implements ObjectSerializer<void> {
  outer: JSONObjectSerializer
  inner: JSONObjectSerializer

  constructor(serializer: JSONSerializer, name: string) {
    this.outer = new JSONObjectSerializer(serializer)

    this.outer.serializeKey('type')
    this.outer.serializeValue(name)

    this.outer.serializeKey('object')
    this.inner = new JSONObjectSerializer(serializer)
  }

  serializeKey<U extends Serializable>(key: U): void {
    return this.inner.serializeKey(key)
  }

  serializeValue<U extends Serializable>(value: U): void {
    return this.inner.serializeValue(value)
  }

  end(): void {
    this.inner.end()
    this.outer.end()
  }
}

export class JSONSerializer implements Serializer<void> {
  output: string = ''

  write(value: string) {
    this.output += value
  }

  serializeString(value: string) {
    this.write(`"${value}"`)
  }

  serializeBoolean(value: boolean): void {
    this.write(value.toString())
  }

  serializeSymbol(value: symbol): void {
    const key = Symbol.keyFor(value)
    if (key) {
      this.write(key)
    } else {
      return this.serializeString(value.toString())
    }
  }

  serializeObject(): ObjectSerializer<void> {
    return new JSONObjectSerializer(this)
  }

  serializeClass(name: PropertyKey): ObjectSerializer<void> {
    return new JSONClassSerializer(this, name as string)
  }

  serializeNumber(value: number) {
    this.write(value.toString())
  }

  serializeBigInt(value: bigint) {
    this.write(value.toString())
  }

  serializeIterable(): IterableSerializer<void> {
    return new JSONIterableSerializer(this)
  }

  serializeNull() {
    return this.write('null')
  }
}


